# -*- coding: utf-8 -*-

from .buffer_logger import BufferLogger
from .cache_utils import cpickle_cache
from .list_utils import ListUtils, uniq_seqs
from .process_notifier import process_notifier
from .speed import Speed
from .string_utils import StringUtils, String
from .items_group_and_indexes import ItemsGroupAndIndexes
from .item_increment_id_dict import ItemIncrementIdDict
from .xml import TextArrayXMLParser
from .dict_utils import DictUtils
from .extract_words import is_nltk_word, extract_words
from .regexp_utils import regexp
from .nlp import is_regular_word, jieba_parse
from .lazy_data import ld
from .unicode_utils import UnicodeUtils, Unicode
from .entropy import calculate_entropy
from .hash_utils import HashUtils
from .itertools_utils import ItertoolsUtils
from .design_pattern import singleton, classproperty, cached_property, environ_property
from .object_utils import set_default_value
from .console_utils import uprint
from .slots_utils import slots_with_pickle
from .json_utils import JsonUtils



# -*- coding: utf-8 -*-

from werkzeug.utils import cached_property, environ_property
from singleton import singleton

class ClassPropertyDescriptor(object):

    def __init__(self, func, name=None, doc=None):
        self.func               = func.__func__
        self.__name__           = name or self.func.__name__
        self.__module__         = self.func.__module__
        self.__doc__            = doc or self.func.__doc__

        # 改用self.is_cached替换ClassPropertyDescriptor, 后者在类上使用装饰器时会失效
        self.is_cached          = False

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)

        if self.is_cached:
            return klass.__dict__.get(self.__name__)
        else:
            value = self.func(klass)
            setattr(klass, self.__name__, value)
            self.is_cached = True
            return value


def classproperty(func):
    """
    Mimic werkzeug.utils's cached_property.

    A decorator that converts a function into a lazy class property.
    The function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @classproperty
            def bar(cls):
                # calculate something important here
                return 42
        Foo.bar # => 42

    The class has to have a `__dict__` in order for this property to
    work.
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)



# -*- coding: utf-8 -*-

from .item_increment_id_dict import ItemIncrementIdDict

class MarkObjectIds(object):
    """
    标记 之前这两个ObjectId是否已经被处理过
    """
    def __init__(self):
        self.mark_object_ids_dict = dict()

    def write(self, oids, value):
        self.mark_object_ids_dict[self.generate_key(oids)] = value

    def read(self, oids):
        k1 = self.generate_key(oids)
        return self.mark_object_ids_dict.has_key(k1) and self.mark_object_ids_dict[k1] or None

    def generate_key(self, oids):
        return ','.join(sorted([str(ItemIncrementIdDict.fetch(oid1)) for oid1 in oids]))



# -*- coding: utf-8 -*-

from .design_pattern import singleton
from collections import defaultdict

@singleton()
class DictUtilsClass(object):
    def nested_read(self, dict1, keys, default_val=None):
        if not isinstance(keys, list): keys = [keys]

        current_dict = dict1
        current_val  = default_val
        for k1 in keys:
            if not isinstance(current_dict, dict): break

            if isinstance(current_dict, dict):
                if current_dict.has_key(k1):
                    current_dict = current_dict[k1]
                    current_val  = current_dict
                else:
                    break

            if isinstance(current_dict, list):
                if len(current_dict) > k1:
                    current_dict = current_dict[k1]
                    current_val  = current_dict
                else:
                    break

        return current_val


    def add_default_value(self, dict1, default_value=None):
        if default_value is None:
            default_value = sum(dict1.values()) / float(len(dict1))
        return defaultdict(lambda : default_value, dict1)

DictUtils = DictUtilsClass()



# -*- coding: utf-8 -*-

import os, sys, time, cPickle
from termcolor import cprint

def cpickle_cache(filename, func1, quiet=False):
    """
    Usage:

    questions = cpickle_cache(pickle_filename, load_questions_func)
    """
    return cPickleCache(filename, func1, quiet).process()


class cPickleCache(object):
    def __init__(self, filename, func1, quiet=False):
        self.filename = filename
        self.func     = func1
        self.quiet    = quiet

    def blue(self, msg, is_ender=False):
        if not self.quiet:
            e1 = "\n" if is_ender else ""
            cprint(msg, 'blue', end=e1)

    def process(self):
        tb = time.time()
        self.blue(u"[%s] " % self.filename)

        if os.path.isfile(self.filename):
            self.blue(u"cache is already exists!")
            result = cPickle.load(open(self.filename, 'rb'))
        else:
            self.blue(u"generating cache ...")
            result = self.func()
            cPickle.dump(result, open(self.filename, 'wb'))
        self.blue(" load in %2.2f sec" % (time.time()-tb), True)
        return result



# -*- coding: utf-8 -*-

__all__ = ['JsonUtils']

import json
from .design_pattern import singleton

# copied from https://github.com/mikexstudios/python-firebase/pull/5/files
import datetime #for parse datetime object to string
import decimal #for parse decimal to string

class JSONEncoder(json.JSONEncoder):
    """ 支持 Decode datetime, decimal等类型 """
    def default(self, obj): # ignore
        if isinstance(obj, datetime.datetime):
            # '2014-10-13T17:51:09.692857'[0:19] => '2014-10-13T17:51:09'
            return obj.isoformat()[0:19]
        elif isinstance(obj, datetime.timedelta):
            return total_seconds(obj)
        elif isinstance(obj, decimal.Decimal):
            return float(obj)
        else:
            return json.JSONEncoder.default(self, obj)

@singleton()
class JsonUtilsClass(object):


    def unicode_dump(self, item1):
        o1 = None # init
        if isinstance(item1, (dict, list, int, bool, float, unicode)):
            o1 = item1
        else:
            if '__dict__' in dir(item1):
                o1 = item1.__dict__
            else:
                raise Exception("%s can't be dumped" % item1)

        # decode("unicode-escape") 无法兼容  " \ 等转义，而是转不回来了
        return json.dumps(o1, cls=JSONEncoder, ensure_ascii=False)

JsonUtils = JsonUtilsClass()



# -*- coding: utf-8 -*-

import os
import progressbar as pb
import humanize


class ProcessNotifier(object):
    def __init__(self, scope, msg):
        """ TypeError: __init__() should return None, not 'generator' """
        from .object_utils import set_default_value

        scope_repr = unicode(repr(type(scope)))

        self.orig_scope = scope
        self.iterator = set_default_value(
              [lambda : scope.iteritems(), lambda : scope.iterator(), lambda: iter(scope), ], \
              scope_repr + u" should be iteratable!")

        self.current_pid = os.getpid()
        self.msg = msg

        # 兼容 list, dict, mongomock, file
        # 判断这个对象的属性方法来觉得用len还是count，不能覆盖所有情况，所以这里直接暴力解决。
        self.total_count = set_default_value( \
            [ lambda : len(scope), lambda : scope.count(), lambda : os.path.getsize(scope.name)], \
              scope_repr + u" should be counted!")
        self.is_file = isinstance(scope, file)


    def generator(self):
        """ Use a Generator to print the inner status when iterate a scope."""
        base_str = (u"%s {pid:%i, records:%i}" % (self.msg, self.current_pid, self.total_count)) \
                   if not self.is_file else \
                   (u"%s {pid:%i, file:%s, size:%s}" % (self.msg, self.current_pid, self.orig_scope.name, humanize.naturalsize(self.total_count)))

        widget_cls = FileProcessSpeed if self.is_file else ItemProcessSpeed
        widgets = [' ', base_str.encode("UTF-8"), pb.Percentage(), ' ', pb.Bar(),
                   ' ', widget_cls()]

        process_count = 0
        pre_process_count = 0
        one_percent = self.total_count / 200

        self.pbar = None
        if self.total_count:
            self.pbar = pb.ProgressBar(widgets=widgets, maxval=self.total_count).start()

        for record in self.iterator:
            process_count += ((len(record) + 1) if self.is_file else 1)
            # Fix grow up when iterator
            if process_count > self.total_count: process_count = self.total_count
            if self.total_count and (process_count - pre_process_count > one_percent):
                self.pbar.update(process_count)
                pre_process_count = process_count
            yield record

        if self.pbar:
            self.pbar.update(process_count) # finish it.
            print


class ItemProcessSpeed(pb.ProgressBarWidget):
    def __init__(self):
        self.fmt = '%6.2f items/s'
    def update(self, pbar):
        if pbar.seconds_elapsed < 2e-6:#== 0:
            bps = 0.0
        else:
            bps = float(pbar.currval) / pbar.seconds_elapsed
        return self.fmt % bps

class FileProcessSpeed(pb.ProgressBarWidget):
    def __init__(self):
        self.fmt = '%s/s'
    def update(self, pbar):
        if pbar.seconds_elapsed < 2e-6:#== 0:
            bps = 0.0
        else:
            bps = float(pbar.currval) / pbar.seconds_elapsed
        return self.fmt % humanize.naturalsize(bps)


def process_notifier(scope, msg=u"",):
    """
    Usage:

    for b1 in process_notifier(a_list):
        # do b1
    """

    assert isinstance(msg, unicode)
    return ProcessNotifier(scope, msg).generator()



# -*- coding: utf-8 -*-

import collections, re
from .regexp_utils import regexp
import math

class StringUtils(object):

    @classmethod
    def merge(self, *strs):
        """ 拼接 包含各种类型对象 的列表里的字符串 """
        result = list()
        for str1 in strs:
            if isinstance(str1, str):
                str1 = str1.decode("utf-8")
            elif isinstance(str1, unicode):
                str1
            else:
                str1 = str(str1)
            result.append(str1)
        return u' '.join(result)


    @classmethod
    def calculate_text_similarity(self, text1, text2, inspect=False, similar_rate_baseline=0.0, skip_special_chars=False):
        """
        目前简单计算 公共字符串比例即可，因为前几步已经算了答案数量和内容都是一样的。

        示例:
        print String.calculate_text_similarity("It\'s ten o\'clock! Let\'s go to bed. Good ______.", "Good ______. ")

        TODO
        误判:
        1. 51e2710da310cdedfd3167d0和51dec4e8a310cdedfd192487 的内容是d[i]ver和dr[i]ve 不同。
        解决，可以判断是仅仅一个单词长度就认为不一样，如果是填字母入单词的话。
        """

        # 如果只是hash就不必计算了，比如 51e9eb90a31047fb0c697413
        if not text1 or not text2 or \
                regexp.object_id.match(text1): # 只要一个match其实就不用比较了
            return {"similarity_rate": 0.0, "info": ""}

        if skip_special_chars:
            text1 = regexp.special_chars.sub("", text1)
            text2 = regexp.special_chars.sub("", text2)

        # 不兼容处理中文等非英文转化为list
        text1_list        = list(text1.strip())
        text2_list        = list(text2.strip())

        original_count    = len(text1_list) + len(text2_list)

        if original_count == 0.0:
            return {"similarity_rate": 0.0, "info": ""}

        # 两两字符串计算重复度用长度较小的那个来做外循环
        t1, t2 = (len(text1_list) > len(text2_list)) and (text2_list, text1_list) or (text1_list, text2_list)
        for char1 in t1[:]: # 使用copy去修复"for循环依赖索引"，感谢罗鑫
            if char1 in t2:
                text1_list.remove(char1)
                text2_list.remove(char1)

        remain_count    = len(text1_list) + len(text2_list)
        similarity_rate = float(original_count - remain_count) / original_count

        if inspect and (similarity_rate > similar_rate_baseline):
            lines = [
                        u"",
                        u"计算相似度详细log",
                        u"-"*40,
                        u"text1: %s" % text1,
                        u"text2: %s" % text2,
                        u"similarity_rate: %f" % similarity_rate,
                        u"-"*40,
                    ]
            info = "\n".join(lines)
        else:
            info = ""

        return {"similarity_rate": similarity_rate, "info": info}


    @classmethod
    def frequence_chars_info(self, str1, length_lambda=lambda len1 : len1):
        # 兼容unicode split
        if isinstance(str1, str): str1 = unicode(str1, "UTF-8")
        str_unicode_len = len(str1)
        default_length  = int(round(length_lambda(math.sqrt(str_unicode_len))))
        result1 = collections.Counter((str1 or "").lower())

        result2 = [t1[1] for t1 in sorted([(-t2, t1) for t1, t2 in result1.most_common()])]

        return {
                "sorted_freq_chars" : ''.join(result2)[0:default_length],
                "uniq_chars__len" : len(result2) }


String = StringUtils



# -*- coding: utf-8 -*-

from datetime import datetime

class Speed():
    def __init__(self):
        self.born_clock = datetime.now()
        self.last_clock = datetime.now()

        self.item_count    = 0

    def tick(self):
        self.last_clock = datetime.now()
        self.item_count += 1
        return self

    def inspect(self):
        if self.item_count == 0:
            print "当前还没有计数"
        else:
            c1 = self.item_count / (self.last_clock - self.born_clock).total_seconds()
            print "当前平均速度为1秒处理", c1, "个"
        return self



# -*- coding: utf-8 -*-

import re
from .lazy_data import ld

def is_nltk_word(str1):
    if not isinstance(str1, unicode):
        try:
            str1 = unicode(str1, "UTF-8")
        except:
            return False # invalid chars
    return (str1 == u'a') or (str1 == u'I') or ((len(str1) >= 2) and (str1 in ld.regular_words))


def extract_words(sentence):
    if isinstance(sentence, list): sentence = ' '.join(sentence)

    sentence = re.compile("[^a-z]+", re.IGNORECASE).sub(" ", sentence)
    these_words = [i1.strip() for i1 in sentence.split(" ")]
    words = [i1 for i1 in these_words if is_nltk_word(i1)]
    return words



# -*- coding: utf-8 -*-

class ItemsGroupAndIndexes(object):
    """
    用于判断 一个元素 在 一个集合中 是否已经和其中某些元素 被group过。
    """

    def __init__(self):
        # 存储 排重结果，类似 [[oi1, oi2], [oi8, oi9, oi10], ...]
        self.result                           = list()
        # 存储 单个题目ID到 [oi1, oi2] 的索引
        self.object_id_to_same_ids_index_dict = dict()

    def exists_between(self, item_id1, item_id2):
        if not self.object_id_to_same_ids_index_dict.has_key(item_id1): return False
        if not self.object_id_to_same_ids_index_dict.has_key(item_id2): return False

        return self.object_id_to_same_ids_index_dict[item_id1] == self.object_id_to_same_ids_index_dict[item_id2]

    def exists(self, item_id):
        return item_id in self.object_id_to_same_ids_index_dict

    def add(self, same_ids):
        for same_id1 in same_ids:
            if self.exists(same_id1):
                continue
            else:
                # 存储数据
                self.result.append(same_ids)
                idx = len(self.result) - 1
                # 存储索引
                for same_id1 in same_ids:
                    self.object_id_to_same_ids_index_dict[same_id1] = idx

    def find(self, item_id):
        idx = self.object_id_to_same_ids_index_dict[item_id]
        return self.result[idx]

    def update(self, item_ids):
        pass

    def groups_len(self) : return len(self.result)
    def items_len(self)  : return len(self.object_id_to_same_ids_index_dict)

    def result_inspect(self):
        for l1 in self.result:
            print ', '.join([str(i1) for i1 in l1])
        print "groups_len", self.groups_len()
        print "items_len", self.items_len()

    def result_json(self):
        return [
                [(((type(i1) in [int, float]) and i1) or str(i1) ) for i1 in r1]
            for r1 in self.result]



# -*- coding: utf-8 -*-

def uprint(*objs):
    """ mimic default print with unicode support """
    text = ', '.join([repr(o1).decode("unicode-escape") for o1 in objs])
    print text
    return text



# -*- coding: utf-8 -*-

import hashlib

from .regexp_utils import regexp
from .design_pattern import singleton

@singleton()
class HashUtilsClass(object):

    def hashvalue_with_sorted(self, str1):
        # sub unicode str
        str1 = regexp.special_chars.sub("", str1)

        # turn into `str` type
        if isinstance(str1, unicode): str1 = str1.encode("UTF-8")

        # sorted in ASNII
        str1 = ''.join(sorted(list(str1)))

        return hashlib.md5(str1).hexdigest()

HashUtils = HashUtilsClass()



# -*- coding: utf-8 -*-

import functools
__all__ = ['slots_with_pickle']


# dont inherit from object, so no __init__ method
class SlotPickleMixin:
    """This mixin makes it possible to pickle/unpickle objects with __slots__ defined.

    In many programs, one or a few classes have a very large number of instances.
    Adding __slots__ to these classes can dramatically reduce the memory footprint
    and improve execution speed by eliminating the instance dictionary. Unfortunately,
    the resulting objects cannot be pickled. This mixin makes such classes pickleable
    again and even maintains compatibility with pickle files created before adding
    __slots__.

    Recipe taken from:
    http://code.activestate.com/recipes/578433-mixin-for-pickling-objects-with-__slots__/
    """
    def __getstate__(self):
        return dict(
            (slot, getattr(self, slot))
            for slot in self.__slots__
            if hasattr(self, slot)
        )

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)


def slots_with_pickle(*slots):
    def func(decorated_class):
        # set slots related
        setattr(decorated_class, '__slots__', slots)
        class _class(SlotPickleMixin, decorated_class): pass

        # set orig class attrs
        _class.__name__    = decorated_class.__name__
        _class.__module__  = decorated_class.__module__
        decorated_class    = _class

        decorated_class = _class
        return decorated_class
    return func
